//--------------------------------------------------
// SkunkNav.js by Greg Kusnick (goric@cognomen.com)
//--------------------------------------------------

//-------------------------------------------------------------------------------
//
// This module contains navigation code for moving your character from place
// to place under script control.  Features include:
//
// *    Define routes with intermediate waypoints.
// *    Include doors and portals in your routes.
// *    Save a route to a file and read it back again.
// *    Import terrain-sensitive routes from AC Explorer Pro.
// *    Supports both run-as-default and walk-as-default.
//
// This version of SkunkNav requires SkunkWorks 1.0 or better.
// See SkunkWorks.html for complete SkunkWorks API documentation.
//
//-------------------------------------------------------------------------------
//
// The principal export of this module is the Route class, which is a sequence
// of map locations (maplocs) to be visited in order on the way to a final
// destination maploc.  Maplocs are obtained from the skapi object using its
// maplocCur property or the MaplocFromLatLng, MaplocFromSz, and MaplocFromArray
// methods, or from any ACO using the aco.maploc proerpty.  See SkunkWorks.html
// for further details on skapi, ACOs and maplocs.
//
// Create an empty route object as follows:
//
//      var route = new Route();
//
// Once created, the route object has the following properties and methods:
//
// route.rgmaploc is an array containing the points on the route.
//
// route.cmaploc is the number of points on the route.
//
// route.imaploc is the index of the waypoint from which to resume an
// interrupted route.
//
// route.Clear() deletes all points from the route and resets it to an empty
// state.
//
// route.AddMaploc(maploc) adds a map location to the end of the route.  Use
// this method to build up your route point by point.
//
// route.AddMaplocCur() adds the logged-in character's current location to the
// end of the route.
//
// route.AddDoor(acoDoor) adds a door to the end of the route.  acoDoor is the
// door to be added, and must be a valid ACO.  If the door is closed when your
// character gets there, Route.Go() will open it.  (If the door is locked, you're
// on your own; there's no provision here for unlocking it.)
//
// route.AddPortal(acoPortal) adds a portal to the end of the route.  acoPortal
// is the portal to be added, and must be a valid ACO.  When your character gets
// there, Route.Go() will portal through and then continue on to the next waypoint
// following the portal.  Note that since portal travel is one-way, using
// route.RouteReverse() on a route containing a portal is not supported.
//
// route1.Append(route2) appends the points in route2 to the end of route1,
// concatenating the two routes.  route2 is not modified by this method (but
// route1 is).
//
// route1.RouteConcat(route2, fOptimize) returns a new route object representing
// the concatenation of route1 with route2.  Neither route1 nor route2 is modified
// by this method.  If fOptimize is false, the two routes are concatenated verbatim.
// If true, redundant points at the juncture of the two routes are pruned away.
// Thus if route2 is the exact reverse of route1, the concatenation will be empty
// (if fOptimize is true).  If only part of route2 matches the reversal of route1,
// then only that portion of both routes will be removed, leaving the unique portions
// intact.  For an example of this, see the comment on the function definition.
//
// route.RouteReverse() returns a new route object with with the order of the
// points reversed.  Use this method to backtrack along a route.  Note that since
// portal travel is one-way, using route.RouteReverse() on a route containing a
// portal is not supported.
//
// route.SaveToFile(szFile) saves the route to the file named by szFile.
//
// route.FromFile(szFile) loads a route from the file named by szFile.  The
// previous contents of the route object are lost.  The specified file can be
// one created by route.SaveToFile, one you've typed in manually, or one you've
// exported from AC Explorer Pro.
//
// The native route file format is a simple list of locations, one per line,
// in standard latitude-and-longitude notation (e.g. 13.645N 0.543E).  To include
// a door or portal, put the name of the object after its location, separated by
// a comma.  Lines beginning with "//" are treated as comments.  Terminate the
// route with a line containing the word "Stop".  See DoorsAndPortals.route and
// Sample.route for examples.
//
// If you're going to use route.FromFile to read ACX Pro route files, make sure
// your ACX Pro terrain data is up to date before calculating and exporting the
// route.  Do this in ACX Pro by choosing "Update terrain..." from the Update
// menu.  Also note that routes calculated in this way are accurate only to
// about 0.1 map unit (about 40 paces).  While they work well on flat or hilly
// terrain, in rugged mountainous regions you may find your character running
// along the bottom of a cliff he's supposed to be on top of.  There's not much
// I can do about this without more precise route information from ACX Pro.
//
// route.Write(ts) saves the route to an open text stream.  Use this method to
// embed routes in other file types, or to save several routes to the same file.
//
// route.Read(ts) loads a route from an open text stream.  The previous contents
// of the route object are lost.  Use this method to read routes embedded in
// other files types, or to read several routes from the same file.
//
// route.Go() moves your character from point to point along the route to the
// endpoint.  Where possible, intermediate waypoints are taken at full running
// speed; where waypoints are close together, your character may slow to a walk,
// or even stop momentarily for sharp turns.  If your character gets stuck in a
// corner or something so that no progress is made for ten seconds, route.Go()
// attempts to back out and try again from a different angle.  (But don't expect
// miracles if he gets stuck indoors.)  When your character arrives within about
// four paces of the route's endpoint, route.Go returns.  To force a premature
// return, manually put your character into (or out of) combat mode.  Also if you
// get killed somewhere along the way, route.Go returns immediately, leaving your
// character standing by the lifestone.
//
// route.StopRoute() stops execution of a route in progress.  Call this from an event
// handler to interrupt route.Go() and return control to the caller of route.Go().
//
// route.ResumeRoute() resumes execution of an interrupted route.
//
// In addition to the Route class and its methods, the following functions are
// exported:
//
// RouteFromMaploc(maploc) creates and returns a direct, straight-line route
// (i.e. a route of just one point) to the given location.
//
// RouteFromFile(szFile) creates and returns a new route object from the contents
// of the named file.
//
// GoToLatLng(lat, lng) moves your character to the given latitude and longitude
// by the most direct route.  Note that the arguments to this function are in
// reverse order compared to those of RunTowardsEastNorth (which takes longitude,
// latitude).
//
// GoToMaploc(maploc) moves your character to the given location by the most
// direct route.  This is equivalent to RouteFromMaploc(maploc).Go().
//
// TurnTowardMaploc(maploc) turns your character to face the given maploc (more
// or less), but doesn't go there.
//
// TurnToHead(head) turns your character to face the given compass heading (in
// degrees, measured clockwise from north).
//
// DistToMaploc(maploc) returns the 2D distance in map units to the given maploc.
//
// Dist3DToMaploc(maploc) returns the 3D distance in map units to the given
// maploc.
//
// HeadToMaploc(maploc) returns the compass heading to the given maploc, i.e.
// the direction you'd have to face to be aimed at that maploc.
//
// DheadToMaploc(maploc) returns the difference in compass heading (the amount
// you would have to turn by) from the current location to the given maploc.
//
// DistMaplocToMaploc(maploc1, maploc2) returns the 2D distance in map units
// from maploc1 to maploc2.
//
// Dist3DMaplocToMaploc(maploc1, maploc2) returns the #D distance in map units
// from maploc1 to maploc2.
//
// HeadMaplocToMaploc(maploc1, mpaloc2) returns the compass heading from maploc1
// to maploc2.
//
// MaplocOfSel() returns the location of the currently selected object in
// maploc form, or null if the selected object has no defined geographical
// location (for instance if it's in your inventory).
//
// UsePortalSel() enters the currently selected portal and waits until your
// character arrives at the other end.
//
//-------------------------------------------------------------------------------
//
// I've also provided the following functions for backward compatibility with
// NavLib.js.  I haven't spent much effort on documenting them because (a) the
// originals in NavLib.js aren't terribly well documented, and (b) you really
// ought to use the native SkunkNav functions above instead.  These
// compatibility functions are provided solely to make it easier to port your
// script from NavLib to SkunkNav.
//
// HeadingTurn(headingNow, headingIntended)
//
// Loc2Coords(landblock)
//
// LocXY2Coords(landblock, x, y)
//
// DebugPosition()
//
// headingXY(xEW_From, yNS_From, xTo, yTo)
//
// distance(xEW, yNS, xEWCur, yNSCur)
//
// HeadingCurrent()
//
// TurnToHeading(headingNew, coefficient)
//
// RunTowardsxEWyNS(xEW, yNS, xEWCur, yNSCur, heading)
//
// GetXEWCur()
//
// GetYNSCur()
//
// StepForward()
//
// SmallJump()
//
// PointToHeading(headingNew)
//
// RunTowardsEastNorth(xEW, yNS)
//
// RunTowardsXY_Ex(xEW, yNS)
//
//-------------------------------------------------------------------------------
//
// This code produces a ton of debug output.  You can reduce that by selectively
// commenting out calls to TraceEnter and TraceExit, or by simply deleting the
// bodies of those functions.
//
//-------------------------------------------------------------------------------

// Tallinu: You can set the following variable at the start of your scripts, or change
// the default setting here to set it for all scripts you run, to disable all navigation
// debug output.

var SkunkNav_Debug = true;


//----------------
// Initialization
//----------------

// Register a handler for OnDeathSelf so we can stop running if we get killed.
var navhandlers = new NavHandlers();
skapi.AddHandler(evidOnDeathSelf, navhandlers);


//--------------------------
// Route class constructor:
//--------------------------
function Route()
{
    this.rgmaploc = new Array();    // Array of points on the route.
    this.cmaploc = 0;               // Number of points on the route.
    this.imaploc = 0;               // Which waypoint to resume from.

    this.Clear = Route_Clear;
    this.AddMaploc = Route_AddMaploc;
    this.AddMaplocCur = Route_AddMaplocCur;
    this.AddDoor = Route_AddDoor;
    this.AddPortal = Route_AddPortal;
    this.Append = Route_Append;
    this.RouteConcat = Route_RouteConcat;
    this.RouteReverse = Route_RouteReverse;
    this.FromFile = Route_FromFile;
    this.SaveToFile = Route_SaveToFile;
    this.Read = Route_Read;
    this.Write = Route_Write;
    this.Load = Route_Load;
    this.Save = Route_Save;
    this.Go = Route_Go;
    this.StopRoute = Route_StopRoute;
    this.ResumeRoute = Route_ResumeRoute;
}


//----------------
// Public methods
//----------------

// Delete all points from the route and reset it to an empty state.
function Route_Clear()
{
    if (SkunkNav_Debug) TraceEnter("Route.Clear", "");

    this.rgmaploc = new Array();
    this.cmaploc = 0;
    this.imaploc = 0;

    if (SkunkNav_Debug) TraceExit("Route.Clear", "");
}

// Add a map location to the end of the route.  Use this method to build up
// your route point by point.
function Route_AddMaploc(maploc)
{
    if (SkunkNav_Debug) TraceEnter("Route.AddMaploc", SzTraceMaploc(maploc));

    this.rgmaploc.push(maploc);
    this.cmaploc++;

    if (SkunkNav_Debug) TraceExit("Route.AddMaploc", "");
}

// Add the logged-in character's current location to the end of the route.
function Route_AddMaplocCur()
{
    if (SkunkNav_Debug) TraceEnter("Route.AddMaplocCur", "");

    this.AddMaploc(skapi.maplocCur);

    if (SkunkNav_Debug) TraceExit("Route.AddMaplocCur", "");
}

// Add a door to the end of the route.  acoDoor is the door to be added, and
// must be a valid ACO.  If the door is closed when your character gets there,
// Route.Go() will open it.  (If the door is locked, you're on your own; there's
// no provision here for unlocking it.)
function Route_AddDoor(acoDoor)
{
    if (SkunkNav_Debug) TraceEnter("Route.AddDoor", SzTraceAco(acoDoor));

    Assert(FAcoIsDoor(acoDoor), "not a door");

    this.AddMaploc(MaplocAcoFromAco(acoDoor));

    if (SkunkNav_Debug) TraceExit("Route.AddDoor", "");
}

// Add a portal to the end of the route.  acoPortal is the portal to be added,
// and must be a valid ACO.  When your character gets there, Route.Go() will
// portal through and then continue on to the next waypoint following the portal.
// Note that since portal travel is one-way, using route.RouteReverse() on a route
// containing a portal is not supported.
function Route_AddPortal(acoPortal)
{
    if (SkunkNav_Debug) TraceEnter("Route.AddPortal", SzTraceAco(acoPortal));

    Assert(FAcoIsPortal(acoPortal), "not a portal");

    this.AddMaploc(MaplocAcoFromAco(acoPortal));

    if (SkunkNav_Debug) TraceExit("Route.AddPortal", "");
}

// Append the points in route2 to the end of this route, concatenating the two
// routes.  route2 is not modified by this method (but this route is).
function Route_Append(route2)
{
if (SkunkNav_Debug) TraceEnter("Route.Append", "");

this.rgmaploc = this.rgmaploc.concat(route2.rgmaploc);
this.cmaploc += route2.cmaploc;

if (SkunkNav_Debug) TraceExit("Route.Append", "");
}

// Return a new route object representing the concatenation of this route with
// route2.  Neither this route nor route2 is modified by this method.  If fOptimize
// is false, the two routes are concatenated verbatim.  If true, redundant points
// at the juncture of the two routes are pruned away.  Thus if route2 is the exact
// reverse of this route, the concatenation will be empty (if fOptimize is true).
// If only part of route2 matches the reversal of this route, then only that portion
// of both routes will be removed, leaving the unique portions intact.
//
// For example, suppose merchant A and merchant B share a shop, or occupy different
// stalls in the same bazaar.  The route in to A will be very similar to the route
// in to B, except at end where the two routes diverge.  The naive way to get from A
// to B would be to concatenate the reverse of route A with route B, like so:
//
//      routeA2B = routeA.RouteReverse().RouteConcat(routeB, false);
//
// While this works, it has the effect of taking your character out of the shop,
// then turning him around and heading right back in to B.  By specifying fOptimize
// as true, this is avoided; the redundant part of the combined route is removed,
// and your character backtracks only to the point where the routes diverge before
// continuing on to B.
function Route_RouteConcat(route2, fOptimize)
{
    if (SkunkNav_Debug) TraceEnter("Route.RouteConcat", "");

    var routeNew = new Route();

    var imaplocFirst1 = 0;
    var imaplocLim1 = this.cmaploc;
    var imaplocFirst2 = 0;
    var imaplocLim2 = route2.cmaploc;

    if (fOptimize)
    {
    //  Try to prune away any pointless backtracking from the juncture of the two
    //  routes.
        while (imaplocLim1 > imaplocFirst1 && imaplocFirst2 < imaplocLim2)
        {
        //  If the last point of route1 coincides with the first point of route2,
        //  omit them both.
            var maploc1 = this.rgmaploc[imaplocLim1 - 1];
            var maploc2 = route2.rgmaploc[imaplocFirst2];
            if (DistMaplocToMaploc(maploc1, maploc2) > distArriveMin)
                break;
            imaplocLim1--;
            imaplocFirst2++;
        }
    }

    var imaploc;
    for (imaploc = imaplocFirst1; imaploc < imaplocLim1; imaploc++)
        routeNew.AddMaploc(this.rgmaploc[imaploc]);
    for (imaploc = imaplocFirst2; imaploc < imaplocLim2; imaploc++)
        routeNew.AddMaploc(route2.rgmaploc[imaploc]);

    if (SkunkNav_Debug) TraceExit("Route.RouteConcat", "");
    return routeNew;
}

// Return a new route object with with order of the points reversed.  Use this
// method to backtrack along a route.  Note that since portal travel is one-way,
// using route.RouteReverse() on a route containing a portal is not supported.
function Route_RouteReverse()
{
    if (SkunkNav_Debug) TraceEnter("Route.RouteReverse", "");

    var routeNew = new Route();
    routeNew.Append(this);
    routeNew.rgmaploc = routeNew.rgmaploc.reverse();

    if (SkunkNav_Debug) TraceExit("Route.RouteReverse", "");
    return routeNew;
}



// Tallinu: These functions allow precise control over files containing multiple routes.
// Lets you read several routes from a single file, or write several routes to a file.
function SkunkNav_OpenFileRead (szFile)
{
    if (SkunkNav_Debug) TraceEnter("SkunkNav_OpenFileRead", SzQuote(szFile));

    if (!SkunkNav_FileSystem)
    {
        SkunkNav_FileSystem = new ActiveXObject("Scripting.FileSystemObject");
    }
    if (SkunkNav_CurrentTextStream) SkunkNav_CurrentTextStream.Close();
    SkunkNav_CurrentTextStream = fso.OpenTextFile (szFile, ForReading);

    if (SkunkNav_Debug) TraceExit("SkunkNav_OpenFileRead", "");
}

function SkunkNav_OpenFileWrite (szFile)
{
    if (SkunkNav_Debug) TraceEnter("SkunkNav_OpenFileWrite", SzQuote(szFile));

    if (!SkunkNav_FileSystem)
    {
        SkunkNav_FileSystem = new ActiveXObject("Scripting.FileSystemObject");
    }
    if (SkunkNav_CurrentTextStream) SkunkNav_CurrentTextStream.Close();
    SkunkNav_CurrentTextStream = fso.OpenTextFile (szFile, ForWriting, true);

    if (SkunkNav_Debug) TraceExit("SkunkNav_OpenFileWrite", "");
}

function SkunkNav_OpenFileAppend (szFile)
{
    if (SkunkNav_Debug) TraceEnter("SkunkNav_OpenFileAppend", SzQuote(szFile));

    if (!SkunkNav_FileSystem)
    {
        SkunkNav_FileSystem = new ActiveXObject("Scripting.FileSystemObject");
    }
    if (SkunkNav_CurrentTextStream) SkunkNav_CurrentTextStream.Close();
    SkunkNav_CurrentTextStream = fso.OpenTextFile (szFile, ForAppending, true);

    if (SkunkNav_Debug) TraceExit("SkunkNav_OpenFileAppend", "");
}

function SkunkNav_CloseFile ()
{
    if (SkunkNav_Debug) TraceEnter("SkunkNav_CloseFile", "");

    if (SkunkNav_CurrentTextStream)
    {
        SkunkNav_CurrentTextStream.Close();
        SkunkNav_CurrentTextStream = null;
    }
    if (SkunkNav_FileSystem) SkunkNav_FileSystem = null;

    if (SkunkNav_Debug) TraceExit("SkunkNav_CloseFile", "");
}

function Route_Load(szTag)
{
    if (SkunkNav_Debug) TraceEnter("Route.Load", "");

    if (SkunkNav_CurrentTextStream) this.Read (SkunkNav_CurrentTextStream, szTag);
    else throw "Error in Route.Load: No file open"

    if (SkunkNav_Debug) TraceExit("Route.Load", "");
}

function Route_Save(szTag)
{
    if (SkunkNav_Debug) TraceEnter("Route.Load", "");

    if (SkunkNav_CurrentTextStream) this.Write (SkunkNav_CurrentTextStream, szTag);
    else throw "Error in Route.Save: No file open"

    if (SkunkNav_Debug) TraceExit("Route.Load", "");
}

// Save the route to the file named by szFile.
// szTag, if passed, will be written as a comment just before the route.
function Route_SaveToFile(szFile, szTag)
{
    if (SkunkNav_Debug) TraceEnter("Route.SaveToFile", SzQuote(szFile));

    var fso = new ActiveXObject("Scripting.FileSystemObject");
    var ts = fso.OpenTextFile(szFile, ForWriting, true);
    this.Write(ts, szTag);
    ts.Close();

    if (SkunkNav_Debug) TraceExit("Route.SaveToFile", "");
}

// Load a route from the file named by szFile.  The previous contents of the
// route object are lost.  The specified file can be one created by
// route.SaveToFile, one you've typed in manually, or one you've exported from
// AC Explorer Pro.  See comments at the top of this module for details of the
// file format.
function Route_FromFile(szFile, szTag)
{
    if (SkunkNav_Debug) TraceEnter("Route.FromFile", SzQuote(szFile));

    var fso = new ActiveXObject("Scripting.FileSystemObject");
    ssay("szfilename= " + szFile);
    var ts = fso.OpenTextFile(szFile, ForReading);
    this.Read(ts, szTag);
    ts.Close();

    if (SkunkNav_Debug) TraceExit("Route.FromFile", "");
}



// Save the route to an open text stream.  Use this method to embed routes in
// other file types, or to save several routes to the same file.
function Route_Write(ts, szTag)
{
    if (SkunkNav_Debug) TraceEnter("Route.Write", "");

    // Tallinu: Added option to write a comment before the route info, can be used to look up a route in a file containing many.
    if (szTag != undefined) ts.WriteLine ("// " + szTag);

    var imaploc;
    for (imaploc = 0; imaploc < this.cmaploc; imaploc++)
    {
        var maploc = this.rgmaploc[imaploc];
        if (maploc.landblock == landblockAco)
        {
        //  Maploc is an object; write out location, name, and ID.
		//	Use semicolon as value separator to avoid localization confusion 
		//	in locales where comma is the decimal separator.
			ts.WriteLine(maploc.sz(3) + ";" + maploc.szAco + ";" + 
              maploc.oidHint.toString(16));
        }
        else
            ts.WriteLine(maploc.sz(3));
    }

    ts.WriteLine("Stop");

    if (SkunkNav_Debug) TraceExit("Route.Write", "");
}



// Tallinu: This skips through a file until it finds a certain comment.
// You must provide the EXACT text that follows the "//", not including leading or trailing spaces.
function SkunkNav_SkipComment (ts, szComment)
{
    while (!ts.AtEndOfStream)
    {
        var szLine = ts.ReadLine();
        if (FPrefixMatchSz(szLine, "//"))
        {
            var szCheck = SzTrimWhiteSpace (SzTail (szLine, "//"));
            if (szCheck == szComment) break;
        }
    }
}


// Load a route from an open text stream.  The previous contents of the route
// object are lost.  Use this method to read routes embedded in other files
// types, or to read several routes from the same file.
function Route_Read(ts, szTag)
{
    if (SkunkNav_Debug) TraceEnter("Route.Read", "");

    if (szTag != undefined) SkunkNav_SkipComment (ts, szTag);
// Tallinu: If szTag is passed, it skips through a file to the commented "tag", then reads the next route.
// Useful for reading a certain route from a file containing a (possibly unknown) number of routes.
// Keeps reading until it finds the comment text specified, or reaches the end of file.
// You must provide the EXACT text that follows the "//", not including leading or trailing spaces.
// Returns an empty route if it doesn't find the specified tag.

    this.Clear();

    while (!ts.AtEndOfStream)
    {
        var szLine = ts.ReadLine();
        var szAco, szMaploc, maploc;
        if (FPrefixMatchSz(szLine, "//"))
        {
        //  Comment line; ignore.
        }
        else if (szLine == "Stop")
        {
            break;
        }
        else if (FPrefixMatchSz(szLine, "!"))
        {
        //  AC Explorer format route file.
            var szKind = SzExtractField(szLine, 2);
            szMaploc = SzExtractField(szLine, 3) + "," +
              SzExtractField(szLine, 4);
            maploc = skapi.MaplocFromSz(szMaploc);
            switch (szKind)
            {
            case "Portal":
            //  REVIEW: Portal name in AC-Spedia DB may not be accurate.
            //  Look for any portal at those coords.
            //  szAco = SzExtractField(szLine, 1)
                this.AddMaploc(MaplocAcoFromMaploc(maploc, ".* Portal", oidNil));
                break;

            case "TempLine":
                this.AddMaploc(maploc);
                var szHead = SzHead(szLine, "=");
                var szTail = SzTail(szLine, "=");
                if (szTail != "")
                {
                    while (szTail != "")
                    {
                        szMaploc = SzHead(szTail, " ");
                        szTail = SzTail(szTail, " ");
                        this.AddMaploc(skapi.MaplocFromSz(szMaploc));
                    }
                }
                else
                {
                    szMaploc = SzExtractField(szLine, 5) + "," +
                      SzExtractField(szLine, 6);
                    this.AddMaploc(skapi.MaplocFromSz(szMaploc));
                }
                break;

            default:
                this.AddMaploc(maploc);
                break;
            }
        }
        
			
	//	Use semicolon as value separator to avoid localization confusion 
	//	in locales where comma is the decimal separator.
		else if (IchInStr(0, szLine, ";") != ichNil)
        {
		//	Maploc is an object; decode location, name, and ID.
			var szT, szOid, oidHint;
			szMaploc = SzHead(szLine, ";");
			szT = SzTail(szLine, ";");
			szAco = SzHead(szT, ";");
			szOid = SzTail(szT, ";");
			TraceLine("szMaploc = " + SzQuote(szMaploc) + ", " + 
			  "szAco = " + SzQuote(szAco) + ", " + 
			  "szOid = " + SzQuote(szOid))
			if (szOid != "")
				oidHint = parseInt(szOid, 16);
			else
				oidHint = oidNil;
			maploc = skapi.MaplocFromSz(szMaploc);
			this.AddMaploc(MaplocAcoFromMaploc(maploc, szAco, oidHint));
        }
        
	//	Accept comma as value separator for compatibility, 
	//	but only if there are periods in the line as well.
		else if (IchInStr(0, szLine, ",") != ichNil && 
		  IchInStr(0, szLine, ".") != ichNil)
        {
        //  Maploc is an object; decode location, name, and ID.
            var szT, szOid, oidHint;
            szMaploc = SzHead(szLine, ",");
            szT = SzTail(szLine, ",");
            szAco = SzHead(szT, ",");
            szOid = SzTail(szT, ",");
            if (SkunkNav_Debug) TraceLine("szMaploc = " + SzQuote(szMaploc) + ", " +
              "szAco = " + SzQuote(szAco) + ", " +
              "szOid = " + SzQuote(szOid))
            if (szOid != "")
                oidHint = parseInt(szOid, 16);
            else
                oidHint = oidNil;
            maploc = skapi.MaplocFromSz(szMaploc);
            this.AddMaploc(MaplocAcoFromMaploc(maploc, szAco, oidHint));
        }
        
        else
        {
            this.AddMaploc(skapi.MaplocFromSz(szLine));
        }
    }

    if (SkunkNav_Debug) TraceExit("Route.Read", "");
}


// Move your character from point to point along the route to the endpoint.
// Where possible, intermediate waypoints are taken at full running speed;
// where waypoints are close together, your character may slow to a walk, or
// even stop momentarily for sharp turns.  If your character gets stuck in a
// corner or something so that no progress is made for ten seconds, route.Go()
// attempts to back out and try again from a different angle.  (But don't expect
// miracles if he gets stuck indoors.)  When your character arrives within about
// four paces of the route's endpoint, route.Go returns.  To force a premature
// return, manually put your character into (or out of) combat mode.  Also if you
// get killed somewhere along the way, route.Go returns immediately, leaving your
// character standing by the lifestone.
function Route_Go()
{
    if (SkunkNav_Debug) TraceEnter("Route.Go", "");

    this.imaploc = 0;
    this.ResumeRoute();

    if (SkunkNav_Debug) TraceExit("Route.Go", "");
}

// Stop execution of a route in progress.  Call this from an event handler to
// interrupt route.Go() and return control to the caller of route.Go().
function Route_StopRoute()
{
    if (SkunkNav_Debug) TraceEnter("Route_StopRoute", "");

    fStop = true;

    if (SkunkNav_Debug) TraceExit("Route_StopRoute", "");
}

// Resume execution of an interrupted route.
function Route_ResumeRoute()
{
    if (SkunkNav_Debug) TraceEnter("Route_ResumeRoute", "");

    DetectRunAsDefault();
    fCombatModeOrig = skapi.fCombatMode;
    fStop = false;

    var timerStuck = skapi.TimerNew();
    var timerLoop = skapi.TimerNew();

    mstate = mstateStop;
    tstate = tstateStraight;
    var maplocPrev = skapi.maplocCur;

    // Visit each waypoint in turn.
    for (; this.imaploc < this.cmaploc; this.imaploc++)
        {
        var maplocTarget = this.rgmaploc[this.imaploc];
        if (SkunkNav_Debug) TraceEnter("Route.Go.OuterLoop",
          this.imaploc.toString() + ": " + SzTraceMaploc(maplocTarget));
        if (FExitGo()) break;

    //  Determine the arrival criterion for this waypoint.  This can vary
    //  from point to point depending on how far apart they are.
        var distArrive = DistArriveFromImaploc(this, this.imaploc);

    //  Keep track of how much progress we're making, so we can tell if
    //  we get stuck.
        var distPrev = 1000;        // Arbitrary large number.
        timerStuck.cmsec = 0;

    //  Loop until we reach the waypoint.
        while (true)
            {
            if (SkunkNav_Debug) TraceEnter("Route.Go.InnerLoop", "");
            fDisableTrace = true;
            if (FExitGo()) break;

        //  Find out where we are and how far it is to the waypoint.
            timerLoop.cmsec = 0;
            var maplocCur = skapi.maplocCur;
            var distCurToTarget = DistMaplocToMaploc(maplocCur, maplocTarget);
            if (SkunkNav_Debug) TraceLine("distCurToTarget = " + distCurToTarget.toFixed(3));

        //  If we've reached this waypoint, move on to the next.
            if (distCurToTarget <= distArrive)
                {
                if (FMaplocIsDoor(maplocTarget))
                    {
                //  Target waypoint is a door, and we're enough to open it.
                fDoorOpen = false
                while (!fDoorOpen)
                 OpenDoorIfClosed(maplocTarget);
                    
                   
                //  Now aim at the corresponding point on the other side of the
                //  door.  This is easy to calculate; it's just door + (door - me),
                //  or 2 * door - me.  This is better than aiming at the door
                //  itself, because if we overshoot there'll be no silly back-
                //  tracking back through the door trying to recover.
                    maplocCur = skapi.maplocCur;
                    maplocTarget = skapi.MaplocFromLatLng(
                      maplocTarget.lat * 2 - maplocCur.lat,
                      maplocTarget.lng * 2 - maplocCur.lng);
                    distArrive = distUseDoor;
                    distPrev = 1000;        // Arbitrary large number.
                    timerStuck.cmsec = 0;
                    distCurToTarget = DistMaplocToMaploc(maplocCur, maplocTarget);
                    }
                else
                    break;
                }

            if (SkunkNav_Debug) TraceLine("distPrev - distCurToTarget = " + (distPrev - distCurToTarget).toFixed(4));
            if (Math.abs(distPrev - distCurToTarget) > 0.005)
                {
            //  Making progress.
                timerStuck.cmsec = 0;
                distPrev = distCurToTarget;
                }
            else
            //  No progress; could be stuck.
                if (SkunkNav_Debug) TraceLine("cmsecStuck = " + timerStuck.cmsec.toString());

            if (timerStuck.cmsec < 10000)
                AdjustCourseAndSpeed(maplocCur, maplocTarget,
                  distCurToTarget, timerLoop);
            else
                {
            //  Haven't made any progress for 10 seconds.  Must be stuck
            //  in a corner or something.  Stop, turn around, and run for
            //  a couple of seconds to try to get unstuck.
                if (SkunkNav_Debug) TraceLine("Stuck.");
                GetUnstuck(maplocCur, maplocPrev);
                maplocTarget = this.rgmaploc[this.imaploc];
                timerStuck.cmsec = 0;
                }

            fDisableTrace = false;
            if (SkunkNav_Debug) TraceExit("Route.Go.InnerLoop", SzTraceMaploc(skapi.maplocCur));
            }
        fDisableTrace = false;
        if (SkunkNav_Debug) TraceExit("Route.Go.InnerLoop", "");    // To cover loop exit.
        if (FExitGo()) break;

        if (FMaplocIsPortal(maplocTarget))
            {
        //  This waypoint is actually a portal.
            UseMaplocPortal(maplocTarget);
            }

        maplocPrev = maplocTarget;
        if (SkunkNav_Debug) TraceExit("Route.Go.OuterLoop", SzTraceMaploc(skapi.maplocCur));
        }
    if (this.imaploc < this.cmaploc)
        if (SkunkNav_Debug) TraceExit("Route.Go.OuterLoop", "");    // To cover loop exit.

    SetMstate(mstateStop);
    SetTstate(tstateStraight);

    // Tallinu - It was only pushing the InventoryPanel key. If that panel was already shown, it would close it! This "panel-dance" makes sure it is opened.
    skapi.Keys(cmidSkillsPanel + cmidInventoryPanel);   // Remove Examine or Help panel by showing Inventory.

    if (SkunkNav_Debug) TraceExit("Route_ResumeRoute", "");
}


//------------------
// Public functions
//------------------

// Create and return a direct, straight-line route (i.e. a route of just one
// point) to the given location.
function RouteFromMaploc(maploc)
{
    if (SkunkNav_Debug) TraceEnter("RouteFromMaploc", SzTraceMaploc(maploc));

    var route = new Route();
    route.AddMaploc(maploc);

    if (SkunkNav_Debug) TraceExit("RouteFromMaploc", "");
    return route;
}

// Create and return a new route object from the contents of the named file.
function RouteFromFile(szFile, szTag)
{
    if (SkunkNav_Debug) TraceEnter("RouteFromFile", SzQuote(szFile));

    var route = new Route();
    route.FromFile(szFile, szTag);

    if (SkunkNav_Debug) TraceExit("RouteFromFile", "");
    return route;
}

// Move your character to the given latitude and longitude by the most direct
// route.  Note that the arguments to this function are in reverse order
// compared to those of RunTowardsEastNorth (which takes longitude, latitude).
function GoToLatLng(lat, lng)
{
    if (SkunkNav_Debug) TraceEnter("GoToMaploc", lat.toFixed(3) + ", " + lng.toFixed(3));

    GoToMaploc(skapi.MaplocFromLatLng(lat, lng));

    if (SkunkNav_Debug) TraceExit("GoToMaploc", "");
}

// Move your character to the given location by the most direct route.  This is
// equivalent to RouteFromMaploc(maploc).Go().
function GoToMaploc(maploc)
{
    if (SkunkNav_Debug) TraceEnter("GoToMaploc", SzTraceMaploc(maploc));

    RouteFromMaploc(maploc).Go();

    if (SkunkNav_Debug) TraceExit("GoToMaploc", "");
}

// Turn your character to face the given maploc (more or less), but don't go there.
function TurnTowardMaploc(maploc)
{
    if (SkunkNav_Debug) TraceEnter("TurnTowardMaploc", SzTraceMaploc(maploc));

    TurnToHead(skapi.maplocCur.HeadToMaploc(maploc));

    if (SkunkNav_Debug) TraceExit("TurnTowardMaploc", "");
}

// Turn your character to face the given compass heading (in degrees, measured
// clockwise from north).
function TurnToHead(head)
{
    if (SkunkNav_Debug) TraceEnter("TurnToHead", head.toFixed(1));

    skapi.TurnToHead(head);
    skapi.WaitEvent(5000, wemSpecific, evidOnActionComplete);

    if (SkunkNav_Debug) TraceExit("TurnToHead", "");
}

// Return the 2D distance in map units to the given maploc.
function DistToMaploc(maploc)
{
    if (SkunkNav_Debug) TraceEnter("DistToMaploc", SzTraceMaploc(maploc));

    var dist = skapi.maplocCur.Dist2DToMaploc(maploc);

    if (SkunkNav_Debug) TraceExit("DistToMaploc", dist.toFixed(3));
    return dist;
}

// Return the 3D distance in map units to the given maploc.
function Dist3DToMaploc(maploc)
{
    if (SkunkNav_Debug) TraceEnter("Dist3DToMaploc", SzTraceMaploc(maploc));

    var dist = skapi.maplocCur.Dist3DToMaploc(maploc);

    if (SkunkNav_Debug) TraceExit("Dist3DToMaploc", dist.toFixed(3));
    return dist;
}

// Return the compass heading to the given maploc, i.e. the direction you'd
// have to face to be aimed at that maploc.
function HeadToMaploc(maploc)
{
    if (SkunkNav_Debug) TraceEnter("HeadToMaploc", SzTraceMaploc(maploc));

    var head = skapi.maplocCur.HeadToMaploc(maploc);

    if (SkunkNav_Debug) TraceExit("HeadToMaploc", head.toFixed(1));
    return head;
}

// Return the difference in compass heading (the amount you would have to turn
// by) from the current location to the given maploc.
function DheadToMaploc(maploc)
{
    if (SkunkNav_Debug) TraceEnter("DheadToMaploc", SzTraceMaploc(maploc));

    var maplocCur = skapi.maplocCur;
    var dhead = DheadNormalize(maplocCur.HeadToMaploc(maploc) -
      maplocCur.head);

    if (SkunkNav_Debug) TraceExit("DheadToMaploc", dhead.toFixed(1))
    return dhead;
}


// Return the 2D distance in map units from maploc1 to maploc2.
function DistMaplocToMaploc(maploc1, maploc2)
{
    if (SkunkNav_Debug) TraceEnter("DistMaplocToMaploc", SzTraceMaploc(maploc1) + ", " + SzTraceMaploc(maploc2));

    var maploc1T, maploc2T;
    if (maploc1.landblock == landblockAco)
        maploc1T = maploc1.maploc;
    else
        maploc1T = maploc1;
    if (maploc2.landblock == landblockAco)
        maploc2T = maploc2.maploc;
    else
        maploc2T = maploc2;

    var dist = maploc1T.Dist2DToMaploc(maploc2T);

    if (SkunkNav_Debug) TraceExit("DistMaplocToMaploc", dist.toFixed(3));
    return dist;
}

// Return the 3D distance in map units from maploc1 to maploc2.
function Dist3DMaplocToMaploc(maploc1, maploc2)
{
    if (SkunkNav_Debug) TraceEnter("Dist3DMaplocToMaploc",
      SzTraceMaploc(maploc1) + ", " + SzTraceMaploc(maploc2));

    var dist = maploc1.Dist3DToMaploc(maploc2);

    if (SkunkNav_Debug) TraceExit("Dist3DMaplocToMaploc", dist.toFixed(3));
    return dist;
}

// Return the compass heading from maploc1 to maploc2.
function HeadMaplocToMaploc(maploc1, maploc2)
{
    if (SkunkNav_Debug) TraceEnter("HeadMaplocToMaploc",
      SzTraceMaploc(maploc1) + ", " + SzTraceMaploc(maploc2));

    var maploc1T, maploc2T;
    if (maploc1.landblock == landblockAco)
        maploc1T = maploc1.maploc;
    else
        maploc1T = maploc1;
    if (maploc2.landblock == landblockAco)
        maploc2T = maploc2.maploc;
    else
        maploc2T = maploc2;

    var head = maploc1T.HeadToMaploc(maploc2T);

    if (SkunkNav_Debug) TraceExit("HeadMaplocToMaploc", head.toFixed(1));
    return head;
}


// Return the location of the currently selected object in maploc form, or null
// if if the selected object has no defined geographical location (for
// instance if it's in your inventory).
function MaplocOfSel()
{
    if (SkunkNav_Debug) TraceEnter("MaplocOfSel", "");

    var aco = skapi.acoSelected;
    var maploc;
    if (aco == null)
        maploc = null;
    else
        maploc = aco.maploc;

    if (SkunkNav_Debug) TraceExit("MaplocOfSel", SzTraceMaploc(maploc));
    return maploc;
}

// Enter the currently selected portal and wait until your character arrives
// at the other end.
function UsePortalSel()
{
    if (SkunkNav_Debug) TraceEnter("UsePortalSel", "");

    Assert(FAcoIsPortal(skapi.acoSelected), "not a portal")

//  Enter the portal.
//  skapi.Keys(cmidUseSelected);
    skapi.acoSelected.Use();

    skapi.WaitEvent(10000, wemSpecific, evidOnStartPortalSelf);
    skapi.WaitEvent(60000, wemSpecific, evidOnEndPortalSelf);

    if (SkunkNav_Debug) TraceExit("UsePortalSel", "");
}


//----------------
// Event handlers
//----------------

function NavHandlers()
{
    this.OnAssessItem = Nav_OnAssessItem;
    this.OnDeathSelf = Nav_OnDeathSelf;
}

function Nav_OnAssessItem(aco, fSuccess, ibi, iai, iwi, iei, ipi)
{
    if (SkunkNav_Debug) TraceEnter("Nav_OnAssessItem", SzTraceAco(aco));

    if (ibi != null)
    {
        fDoorOpen = ibi.fOpen;
        if (aco.szname !="Alembic") ssay("Door is open " + fDoorOpen);
        if (SkunkNav_Debug) TraceLine("Door is " + (fDoorOpen ? "open" : "closed") + ".");
    }
    else
        if (SkunkNav_Debug) TraceLine("No Item Basic Information.");

    if (SkunkNav_Debug) TraceExit("Nav_OnAssessItem", "");
}

function Nav_OnDeathSelf(acoKiller, szMsg)
{
    if (SkunkNav_Debug) TraceEnter("Nav_OnDeathSelf", SzQuote(szMsg));

    fStop = true;

    if (SkunkNav_Debug) TraceExit("Nav_OnDeathSelf", "");
}


//----------------------------------------------------------------------------
// NavLib.js compatibility functions.
// These functions are provided solely to make it easier to port your script
// from NavLib.js to SkunkNav.js.  You really ought to use the native
// SkunkNav functions above instead.
//----------------------------------------------------------------------------

function HeadingTurn(headingNow, headingIntended)
{
    return DheadNormalize(headingIntended - headingNow)
}

function Loc2Coords(landblock)
{
//  Does not work indoors!
    var x = ((landblock >> 3) & 0x7) * 24;
    var y = ((landblock >> 0) & 0x7) * 24;
    return LocXY2Coords(landblock, x, y);
}

function LocXY2Coords(landblock, x, y)
{
    var lat = ((((landblock >> 16) & 0xFF) - 0x7F) * 192 + y - 84) / 240;
    var lng = ((((landblock >> 24) & 0xFF) - 0x7F) * 192 + x - 84) / 240;
    return new Array(lat, lng);
}

function DebugPosition()
{
    ac.Debug(SzTraceMaploc(skapi.maplocCur) + "\n");
}

function headingXY(xEW_From, yNS_From, xTo, yTo)
{
    var dx = xTo - xEW_From;
    var dy = yTo - yNS_From;
    var headRad = Math.atan2(dx, dy);
    if (headRad < 0) headRad += 2 * Math.PI;
    return headRad * 180/Math.PI;
}

function distance(xEW, yNS, xEWCur, yNSCur)
{
    return Math.sqrt(Math.pow(xEWCur - xEW, 2) + Math.pow(yNSCur - yNS, 2));
}

function HeadingCurrent()
{
    return skapi.maplocCur.head;
}

function TurnToHeading(headingNew, coefficient)
{
    return TurnToHead(headingNew);
}

function RunTowardsxEWyNS(xEW, yNS, xEWCur, yNSCur, heading)
{
    GoToLatLng(yNS, xEW);
}

function GetXEWCur()
{
    return skapi.maplocCur.lng;
}

function GetYNSCur()
{
    return skapi.maplocCur.lat;
}

function StepForward()
{
    skapi.Keys(cmidHoldRun, kmoDown)
    skapi.Keys(cmidWalkForward, kmoDown);
//  skapi.Sleep(100);
    skapi.Keys(cmidWalkForward, kmoUp);
    skapi.Keys(cmidHoldRun, kmoUp)
}

function SmallJump()
{
    skapi.Keys(cmidJump, kmoDown);
    skapi.Sleep(100);
    skapi.Keys(cmidJump, kmoUp);
}

function PointToHeading(headingNew)
{
    return TurnToHead(headingNew);
}

function RunTowardsEastNorth(xEW, yNS)
{
    GoToLatLng(yNS, xEW);
}

function RunTowardsXY_Ex(xEW, yNS)
{
    GoToLatLng(yNS, xEW);
}


//---------------------------------
// Private variables and constants
//---------------------------------

var cyardPerMU = 240;           // One map unit = 240 yards.

var distWalk = 0.030;           // Walk when we get this close (~30 paces).
var distArriveRun = 0.015;      // Next waypoint when we get this close (~10 paces).
var distArriveMin = 0.0075;     // Stop when we get this close (~3 paces).
var distArrivePortal = 0.15;    // Use portal when we get this close (~60 paces).
var distArriveDoor = 0.02;      // Open a door when we get this close (~8 paces).
var distUseDoor = 0.005;        // Aim this close to go through a door (~1 pace).
var vWalk = 0.01;               // Walking speed in MU/s.
var vRun = 0.035;               // Typical running speed in MU/s.
var dheadPerSec = 90;           // Turning speed in degrees/s.
var dheadPerSecRun = 133;       // Turning speed in degrees/s while running.
var dheadRadPerSec = 1.57;      // Turning speed in radians/s.
var dheadRadPerSecRun = 2.32;   // Turning speed in radians/s while running.

var cmsecMaxInterval = 1000;    // Adjust course at least once a second.
var cmsecMinInterval = 100;     // Can't adjust any faster than this.

var dheadMinTurn = 3;           // Can't steer more accurately than this.

// Motion state:
var mstateStop = 0;
var mstateWalk = 1;
var mstateRun = 2;
var mstate = mstateStop;

// Turning state:
var tstateLeft = -1;
var tstateStraight = 0;
var tstateRight = 1;
var tstate = tstateStraight;

// For file system calls:
var ForReading = 1;
var ForWriting = 2;
var ForAppending = 8;
var SkunkNav_FileSystem = null;
var SkunkNav_CurrentTextStream = null;


var fRunAsDefault;


//-------------------
// Private functions
//-------------------

// Detect the user's "Run as Default Movement" setting (on AC's Character options
// panel) so we know which key combos to use for walking and running.
function DetectRunAsDefault()
{
    if (SkunkNav_Debug) TraceEnter("DetectRunAsDefault", "");

    fRunAsDefault = ((skapi.chop & chopRunAsDefault) != 0);
    if (SkunkNav_Debug) TraceLine("fRunAsDefault = " + fRunAsDefault.toString());

    if (SkunkNav_Debug) TraceExit("DetectRunAsDefault", "");
}


// Test skapi.fCombatMode to see if user wants to break off navigation.
// Also if we died, stop moving and stay by the lifestone.
function FExitGo()
{
    if (SkunkNav_Debug) TraceEnter("FExitGo", "");

//  Hack to let user break out if things go haywire.
    var fResult;
    if (skapi.fCombatMode != fCombatModeOrig || fStop)
        fResult = true;
    else
        fResult = false;

    if (SkunkNav_Debug) TraceExit("FExitGo", fResult.toString());
    return fResult;
}

var fCombatModeOrig, fStop;


// Calulate an arrival criterion for the specific waypoint.  The idea is
// that when points are far apart (as when running across open country),
// we needn't obsess about hitting the intermediate waypoints dead on; the
// important thing is to keep moving and not slow down.  But when the points
// are close together (as when entering or leaving a shop), we'd better slow
// down and hit them as accurately as we can.
function DistArriveFromImaploc(route, imaploc)
{
    if (SkunkNav_Debug) TraceEnter("Route.DistArriveFromImaploc", imaploc.toString());

    var maplocThis = route.rgmaploc[imaploc];

//  Calculate distances to adjacent waypoints.
    var maplocPrev, distPrevToThis;
    if (imaploc == 0)
    //  No previous waypoint; use player's current position.
        maplocPrev = skapi.maplocCur;
    else
        maplocPrev = route.rgmaploc[imaploc - 1];
    distPrevToThis = DistMaplocToMaploc(maplocPrev, maplocThis);

    var maplocNext, distThisToNext;
    if (imaploc + 1 < route.cmaploc)
    {
        maplocNext = route.rgmaploc[imaploc + 1];
        distThisToNext = DistMaplocToMaploc(maplocThis, maplocNext);
    }
    else
    //  No next waypoint; go for max accuracy at endpoint.
        distThisToNext = 0;

    var distArrive;
    if (FMaplocIsDoor(maplocThis))
        distArrive = distArriveDoor;
    else if (FMaplocIsPortal(maplocThis))
    //  Give portal lots of leeway so we don't run through it by accident.
        distArrive = distArrivePortal;
    else if (distPrevToThis >= distWalk && distThisToNext >= distWalk)
        distArrive = distWalk;
    else
        distArrive = distArriveMin;

    if (SkunkNav_Debug) TraceExit("Route.DistArriveFromImaploc", distArrive.toFixed(3));
    return distArrive;
}

// See if we're on on course for maplocTarget and correct as necessary.
function AdjustCourseAndSpeed(maplocCur, maplocTarget,
  distCurToTarget, timerLoop)
{
    if (SkunkNav_Debug) TraceEnter("AdjustCourseAndSpeed",
      SzTraceMaploc(maplocCur) + ", " + SzTraceMaploc(maplocTarget) + ", " +
      distCurToTarget.toFixed(3));

    var headTarget = HeadMaplocToMaploc(maplocCur, maplocTarget);
    var dhead = DheadNormalize(headTarget - maplocCur.head);
    var mstateNew;
    if (Math.abs(dhead) > 90)
        mstateNew = mstateStop;     // Stop for sharp turns.
    else if (distCurToTarget <= distWalk)
        mstateNew = mstateWalk;
    else
        mstateNew = mstateRun;

//  Do the turn at the slower of the old & new speeds.
    if (mstateNew < mstate) SetMstate(mstateNew);

    if (Math.abs(dhead) > dheadMinTurn)
        TurnToHead(headTarget);

//  Time spent calculating doesn't count while we're stopped.  Start timing from
//  the moment we start moving.
    if (mstate == mstateStop) timerLoop.cmsec = 0;

    if (mstateNew > mstate) SetMstate(mstateNew);

//  Figure out how long until we reach the target at this speed.
    var v = VFromMstate(mstate);
    var cmsec;
    if (v > 0)
    {
    //  Divide by two so we can re-evaluate when we're halfway there.
        cmsec = Math.round((distCurToTarget/v * 1000) / 2);
    //  Account for the time we've already spent calculating.
        cmsec = cmsec - timerLoop.cmsec;
    }
    else
        cmsec = 100000; // Arbitrary large number.

    if (cmsec < cmsecMinInterval)
    {
    //  Calculated time is too short to measure reliably; just tap the key.
        SetMstate(mstateStop);
    //  Give WaitEvent some processing time.
        skapi.WaitEvent(cmsecMinInterval);
    }
    else
    {
        if (cmsec > cmsecMaxInterval)
            cmsec = cmsecMaxInterval;
        skapi.WaitEvent(cmsec, wemFullTimeout);
    }

    if (SkunkNav_Debug) TraceExit("AdjustCourseAndSpeed", "");
}

// Estimate our current traveling speed in map units/sec.
function VFromMstate(mstate)
{
    switch (mstate)
    {
    case mstateStop:
        return 0;
    case mstateWalk:
        return vWalk;
    case mstateRun:
        return vRun;
    }
}


// Turn and run in a semi-random direction for a few seconds.
function GetUnstuck(maplocCur, maplocPrev)
{
    if (SkunkNav_Debug) TraceEnter("GetUnstuck", "");

    SetMstate(mstateStop);
    SetTstate(tstateStraight);

//  Head roughly back the way we came, but with a bit of randomness.
    var head = HeadMaplocToMaploc(maplocCur, maplocPrev);
    head += Math.random() * 90 - 45;
    TurnToHead(head);

    SetMstate(mstateRun);
    skapi.WaitEvent(3000, wemFullTimeout);
    SetMstate(mstateStop);

    if (SkunkNav_Debug) TraceExit("GetUnstuck", "");
}


// Set our state of motion (running, walking, or stopped) to mstateNew.
function SetMstate(mstateNew)
{
    if (SkunkNav_Debug) TraceEnter("SetMstate", mstate.toString() + ", " + mstateNew.toString());

    switch (mstate * 4 + mstateNew)
    {
    case mstateStop * 4 + mstateWalk:
        if (fRunAsDefault) skapi.Keys(cmidHoldRun, kmoDown);
        skapi.Keys(cmidWalkForward, kmoDown);
        break;

    case mstateWalk * 4 + mstateStop:
        skapi.Keys(cmidWalkForward, kmoUp);
        if (fRunAsDefault) skapi.Keys(cmidHoldRun, kmoUp);
        break;

    case mstateStop * 4 + mstateRun:
        if (!fRunAsDefault) skapi.Keys(cmidHoldRun, kmoDown);
        skapi.Keys(cmidWalkForward, kmoDown);
        break;

    case mstateRun * 4 + mstateStop:
        skapi.Keys(cmidWalkForward, kmoUp);
        if (!fRunAsDefault) skapi.Keys(cmidHoldRun, kmoUp);
        break;

    case mstateWalk * 4 + mstateRun:
        if (fRunAsDefault)
            skapi.Keys(cmidHoldRun, kmoUp);
        else
            skapi.Keys(cmidHoldRun, kmoDown);
        break;

    case mstateRun * 4 + mstateWalk:
        if (fRunAsDefault)
            skapi.Keys(cmidHoldRun, kmoDown);
        else
            skapi.Keys(cmidHoldRun, kmoUp);
        break;
    }

    mstate = mstateNew

    if (SkunkNav_Debug) TraceExit("SetMstate", "");
}

// Set our state of turning (left, right, or straight ahead) to tstateNew.
function SetTstate(tstateNew)
{
    if (SkunkNav_Debug) TraceEnter("SetTstate", tstate.toString() + ", " + tstateNew.toString());

    if (tstate != tstateNew)
    {
        switch (tstate)
        {
        case tstateLeft:
            skapi.Keys(cmidTurnLeft, kmoUp);
            break;
        case tstateRight:
            skapi.Keys(cmidTurnRight, kmoUp)
            break;
        }

        tstate = tstateNew;

        switch (tstate)
        {
        case tstateLeft:
            skapi.Keys(cmidTurnLeft, kmoDown);
            break;
        case tstateRight:
            skapi.Keys(cmidTurnRight, kmoDown);
            break;
        }
    }

    if (SkunkNav_Debug) TraceExit("SetTstate", "");
}


// // Given our current position and heading in maplocCur, turn to face maploc.
// function TurnTowardMaplocEx(maploc, maplocCur)
//  {
//  if (SkunkNav_Debug) TraceEnter("TurnTowardMaplocEx",
//    SzTraceMaploc(maploc) + ", " + SzTraceMaploc(maplocCur));
//
//  TurnToHeadEx(HeadMaplocToMaploc(maplocCur, maploc), maplocCur);
//
//  if (SkunkNav_Debug) TraceExit("TurnTowardMaplocEx", "");
//  }

// // Given our current position and heading in maplocCur, turn to the compass
// // heading specified by head.
// function TurnToHeadEx(head, maplocCur)
//  {
//  if (SkunkNav_Debug) TraceEnter("TurnToHeadEx",
//    head.toFixed(1) + ", " + maplocCur.head.toFixed(1));
//
//  TurnByDhead(DheadNormalize(head - maplocCur.head));
//
//  if (SkunkNav_Debug) TraceExit("TurnToHeadEx", "");
//  }

// Normalize the given turning angle to within plus-or-minus 180 degrees.
function DheadNormalize(dhead)
{
    if (SkunkNav_Debug) TraceEnter("DheadNormalize", dhead.toFixed(1));

    var dheadT = dhead;
    while (dheadT > 180)
        dheadT = dheadT - 360;
    while (dheadT < -180)
        dheadT = dheadT + 360;

    if (SkunkNav_Debug) TraceExit("DheadNormalize", dheadT.toFixed(1));
    return dheadT;
}

// // Turn by the specified angle (in degrees).
// function TurnByDhead(dhead)
//  {
//  if (SkunkNav_Debug) TraceEnter("TurnByDhead", dhead.toFixed(1));
//
//  var cmsec;
//  if (mstate == mstateRun ||
//    (mstate == mstateStop && fRunAsDefault))
//      cmsec = Math.round(Math.abs(dhead)/dheadPerSecRun * 1000);
//  else
//      cmsec = Math.round(Math.abs(dhead)/dheadPerSec * 1000);
//  if (SkunkNav_Debug) TraceLine("mstate = " + mstate.toString() + ", " +
//    "cmsec = " + cmsec.toString());
//
//  if (dhead >= dheadMinTurn)
//      {
//      skapi.Keys(cmidTurnRight, kmoDown);
//      skapi.Sleep(cmsec);
//      skapi.Keys(cmidTurnRight, kmoUp);
//      }
//  else if (dhead <= -dheadMinTurn)
//      {
//      skapi.Keys(cmidTurnLeft, kmoDown);
//      skapi.Sleep(cmsec);
//      skapi.Keys(cmidTurnLeft, kmoUp);
//      }
//
//  if (SkunkNav_Debug) TraceExit("TurnByDhead", "");
//  }



// Bogus landblock value to identify a door or portal stored in a route.
var landblockAco = 0xFFFFABCD;

// Phony maploc class for including door and portals in routes.
function MaplocAco(maploc, szAco, oidHint)
{
    this.maploc = maploc;
    this.szAco = szAco;
    this.oidHint = oidHint;
    this.landblock = landblockAco;
    this.lat = maploc.lat;
    this.lng = maploc.lng;
    this.sz = MaplocAco_sz;
}

// sz method for MaplocAco.  Just pass through to the embedded maploc.
function MaplocAco_sz(cdigit)
{
    return this.maploc.sz(cdigit);
}

// Construct a pseudo-maploc to represent a door or portal.
function MaplocAcoFromAco(aco)
{
    if (SkunkNav_Debug) TraceEnter("MaplocAcoFromAco", SzTraceAco(aco));

    var maplocAco = MaplocAcoFromMaploc(aco.maploc,
      aco.szName, aco.oid);

    if (SkunkNav_Debug) TraceExit("MaplocAcoFromAco", SzTraceMaploc(maplocAco));
    return maplocAco;
}

// Construct a pseudo-maploc to represent an object at maploc named szAco.
function MaplocAcoFromMaploc(maploc, szAco, oidHint)
{
    if (SkunkNav_Debug) TraceEnter("MaplocAcoFromMaploc",
      SzTraceMaploc(maploc) + ", " + SzQuote(szAco) + ", " + oidHint.toString(16));

    var maplocAco = new MaplocAco(maploc, szAco, oidHint);

    if (SkunkNav_Debug) TraceExit("MaplocAcoFromMaploc", SzTraceMaploc(maplocAco));
    return maplocAco;
}


// Return true if the given maploc represents a door.
function FMaplocIsDoor(maploc)
{
    if (SkunkNav_Debug) TraceEnter("FMaplocIsDoor", SzTraceMaploc(maploc));

    var fResult = maploc.landblock == landblockAco &&
      FSuffixMatchSz(maploc.szAco, "Door");

    if (SkunkNav_Debug) TraceExit("FMaplocIsDoor", fResult.toString());
    return fResult;
}

// Return true if the given object is a door.
function FAcoIsDoor(aco)
{
    if (SkunkNav_Debug) TraceEnter("FAcoIsDoor", SzTraceAco(aco));

    var fResult = (aco.oty & otyDoor) != 0;

    if (SkunkNav_Debug) TraceExit("FAcoIsDoor", fResult.toString());
    return fResult;
}

// Assess the given door to see if it's open, and if not, open it.
function OpenDoorIfClosed(maplocDoor)
{
    fDisableTrace = false;
    if (SkunkNav_Debug) TraceEnter("OpenDoorIfClosed", SzTraceMaploc(maplocDoor));

//  Don't walk through door until we're sure it's open.
    SetMstate(mstateStop);

//  Find the door and select it.
    var acoDoor = AcoFindNearestSz(maplocDoor.szAco, maplocDoor.oidHint, maplocDoor.maploc); // Tallinu: added last parameter for searching
    if (acoDoor == null)
        throw "OpenDoorIfClosed: Can't find " + maplocDoor.szAco + ".";
    skapi.SelectAco(acoDoor);

//  Process queued events to get any pending assessments out of the way.
    skapi.WaitEvent(0);

//  Tallinu: Modified this section to examine door before using it again, to avoid closing what we may have just opened, if the server doesn't update our assessment.
    fDoorOpen = false;
    skapi.AddHandler(evidOnAssessItem, navhandlers);
    //skapi.outputline("Done Assessing Door(1)" + fDoorOpen);
    
//  Wait for the results.
    var itry;
    for (itry = 0; itry < 5; itry++)
    {
        skapi.WaitEvent(0);
	    skapi.assessaco(alem)
	    skapi.WaitEvent(10000, wemSpecific, evidOnAssessItem);
		skapi.WaitEvent(0);
    
		skapi.assessaco(acoDoor)
		skapi.WaitEvent(10000, wemSpecific, evidOnAssessItem);
		if (fDoorOpen) break;
		if (skapi.acoselected != null)  skapi.acoselected.use();
		skapi.waitevent(1000,wemFullTimeout);
    }
    skapi.RemoveHandler(evidOnAssessItem, navhandlers);

//  Remove the Assess panel if present.
    skapi.Keys(cmidOptionsPanel + cmidInventoryPanel);

    if (SkunkNav_Debug) TraceExit("OpenDoorIfClosed", "");
    fDisableTrace = true;
}

var fDoorOpen;


// Return true if the given maploc represents a portal.
function FMaplocIsPortal(maploc)
{
    if (SkunkNav_Debug) TraceEnter("FMaplocIsPortal", SzTraceMaploc(maploc));

    var fResult = maploc.landblock == landblockAco &&
      FSuffixMatchSz(maploc.szAco, "Portal");

    if (SkunkNav_Debug) TraceExit("FMaplocIsPortal", fResult.toString());
    return fResult;
}

// Return true if the given object is a portal.
function FAcoIsPortal(aco)
{
    if (SkunkNav_Debug) TraceEnter("FAcoIsPortal", SzTraceAco(aco));

    var fResult = (aco.oty & otyPortal) != 0;

    if (SkunkNav_Debug) TraceExit("FAcoIsPortal", fResult.toString());
    return fResult;
}

// Enter the portal and wait until we arrive at the other end.
function UseMaplocPortal(maplocPortal)
{
    if (SkunkNav_Debug) TraceEnter("UseMaplocPortal", SzTraceMaploc(maplocPortal));

//  Find the portal and select it.
    var acoPortal = AcoFindNearestSz(maplocPortal.szAco, maplocPortal.oidHint);
    if (acoPortal == null)
        throw "UseMaplocPortal: Can't find " + maplocPortal.szAco + ".";
    skapi.SelectAco(acoPortal);

    SetMstate(mstateStop);

    UsePortalSel();

    if (SkunkNav_Debug) TraceExit("UseMaplocPortal", "");
}


// Find the nearest object with the specified name (or something like it).
// If oidHint is valid, use it as a hint to find the object faster.
function AcoFindNearestSz(sz, oidHint, targetMaploc)
{
    if (SkunkNav_Debug) TraceEnter("AcoFindNearestSz", SzQuote(sz) + ", " + oidHint.toString(16));

    var acoFound = null;

    var rex = new RegExp("^" + sz + "$", "i");

    var aco;

//  Try oidHint first.
    aco = skapi.AcoFromOid(oidHint)
    if (aco != null)
    {
        skapi.SelectAco(aco);
        aco = skapi.acoSelected;
        if (aco != null)
        {
            if (SkunkNav_Debug) TraceLine(SzTraceAco(aco));
        //  Check name to make sure OID hasn't changed.
            if (rex.test(aco.szName))
                acoFound = aco;
        }
    }

    // Tallinu: Try to find a matching item at the correct location
    if (acoFound == null && targetMaploc != undefined)
    {
        acoFound = GetItemNearMaploc (sz, targetMaploc);
    }

    if (acoFound == null)
    {
    //  oidHint wasn't valid; use \ and ] to search nearby objects.
        skapi.Keys(cmidClosestItem);
        skapi.Sleep(200);
        aco = skapi.acoSelected;
        if (aco != null)
        {
            var oidFirst = aco.oid;
            var ctry = 0;
            while (true)
            {
                if (SkunkNav_Debug) TraceLine(SzTraceAco(aco));
                if (rex.test(aco.szName))
                {
                    acoFound = aco;
                    break;
                }
                if (++ctry == 20) break;
                if (ctry > 2)
                {
                //  Stop moving so we don't overrun it while we're searching.
                    SetMstate(mstateStop);
                }
                skapi.Keys(cmidNextItem);
                skapi.Sleep(200);
                aco = skapi.acoSelected;
                if (aco.oid == oidFirst) break;
            }
        }
    }

    if (SkunkNav_Debug) TraceExit("AcoFindNearestSz", SzTraceAco(acoFound));
    return acoFound;
}

// Function to find an item near a certain maploc. Much better than the \ and ] keys!
function GetItemNearMaploc (findname, targetMaploc)
{
    var foundObject = null; // Nothing found yet!
    var foundObjectDistance = 0; // Has to be closer than this.

    // Pick up a list of all portals and "equipment", which includes doors.
    var ItemCollection = CoacoFromOcm (ocmPortal & ocmEquipment);

    for (var index = 0; index < ItemCollection.Count; index++)
    {
        var currentitem = ItemCollection.Item(index);
        if (currentitem.szName == findname && currentitem.maploc) // if name matches and item has a defined location
        {
            // Get the item's location
            var location = currentitem.maploc

            // Calculate distance between object and target location
            var dist = Dist3DMaplocToMaploc (targetMaploc, location);

            if (dist < foundObjectDistance || foundObject == null)
            {
                foundObject = currentitem;
                foundObjectDistance = dist;
            }
        }
    }

    // Returns null if nothing found, otherwise the object closest to the specified coordinates.
    return foundObject;
}




/////////////////////////////////////////




// Dummy out these functions to reduce the debug output clutter.

var fDisableTrace = false;

function TraceEnter(szSub, szArgs)
{
    if (fDisableTrace) return;

    skapi.OutputLine("[ " + szSub + "(" + szArgs + ")", opmDebugLog);
}

function TraceExit(szSub, szResult)
{
    if (fDisableTrace) return;

    skapi.OutputLine("] " + szSub + ": " + szResult, opmDebugLog);
}

function TraceLine(sz)
{
    if (fDisableTrace) return;

    skapi.OutputLine("  " + sz, opmDebugLog);
}

function SzTraceMaploc(maploc)
{
    if (maploc == null)
        return "null";
    else
        return maploc.sz(3);
}

function SzTraceAco(aco)
{
    if (aco == null)
        return "null";
    else
        return SzQuote(aco.szName);
}


//----------------------------------
// General-purpose string functions
//----------------------------------

var ichNil = -1;

// Return true iff szPrefix matches the beginning (or all) of sz.
// The comparison is case-insensitive.
function FPrefixMatchSz(sz, szPrefix)
{
    var cchSz = sz.length;
    var cchPrefix = szPrefix.length;
    return cchPrefix <= cchSz &&
      sz.substr(0, cchPrefix).toLowerCase() == szPrefix.toLowerCase();
}

// Return true iff szSuffix matches the end (or all) of sz.
// The comparison is case-insensitive.
function FSuffixMatchSz(sz, szSuffix)
{
    var cchSz = sz.length;
    var cchSuffix = szSuffix.length;
    return cchSuffix <= cchSz &&
      sz.substr(cchSz - cchSuffix).toLowerCase() == szSuffix.toLowerCase();
}

// Return the zero-origin position of the first instance (if any) of sz2 within
// sz1, starting at position ichFirst or later.  If not found, return -1.
function IchInStr(ichFirst, sz1, sz2)
{
    return sz1.indexOf(sz2, ichFirst);
}

// Search sz for the first occurrence of szSep and return the portion of sz prior to
// szSep.  If szSep is not found, return all of sz.
function SzHead(sz, szSep)
{
    var ich = IchInStr(0, sz, szSep);
    if (ich >= 0)
        return sz.substr(0, ich);
    else
        return sz;
}

// Search sz for the first occurrence of szSep and return the portion following szSep.
// If szSep is not found, return the empty string.
function SzTail(sz, szSep)
{
    var ich = IchInStr(0, sz, szSep);
    if (ich >= 0)
    {
        var szTail = sz.substr(ich + szSep.length);
        if (szSep == " ")
            szTail = SzTrimWhiteSpace(szTail);
        return szTail;
    }
    else
        return "";
}

// Extract the (zero-origin) ifield'th field from a line of comma-separated fields.
function SzExtractField(szLine, ifield)
{
    var ifieldT = 0;
    var ichFirst = 0;
    var ichLim;
    while (true)
    {
        ichLim = IchInStr(ichFirst, szLine, ",");
        if (ichLim == -1)
        {
            if (ifieldT < ifield)
                return "";
            ichLim = szLine.length;
        }
        if (ifieldT == ifield)
            break;
        ifieldT++;
        ichFirst = ichLim + 1;
    }

    return SzTrimWhiteSpace(szLine.substr(ichFirst, ichLim - ichFirst));
}

// Return the given string wrapped in double-quotes.
function SzQuote(sz)
{
    return "\"" + sz + "\"";
}

// Return a copy of sz with leading and trailing spaces and tabs removed.
function SzTrimWhiteSpace(sz)
{
    var ichFirst, ichLim;
    for (ichFirst = 0; ichFirst < sz.length; ichFirst++)
        if (sz.substr(ichFirst, 1) != " " &&
           sz.substr(ichFirst, 1) != "\t")
            break;

    for (ichLim = sz.length; ichLim > ichFirst; ichLim--)
        if (sz.substr(ichLim - 1, 1) != " " &&
           sz.substr(ichLim - 1, 1) != "\t")
            break;

    return sz.substr(ichFirst, ichLim - ichFirst);
}


// Raise an error if the specified Boolean condition f is not true.
// sz is a string describing the condition for error-reporting purposes.
function Assert(f, sz)
{
    if (!f)
    {
        if (SkunkNav_Debug) TraceLine("Assertion failure: " + sz);
        throw "Assert: " + sz;
    }
}
